/*-------------------------------------------------------------------------+
| Copyright 2022 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.tooling.ext.reuse.utils;

import static org.eclipse.emf.ecore.util.EcoreUtil.delete;
import static org.eclipse.emf.ecore.util.EcoreUtil.replace;
import static org.fortiss.tooling.ext.reuse.ToolingReuseActivator.getDefault;
import static org.fortiss.tooling.ext.reuse.storage.ReuseLibraryStorageManager.deleteReuseLibraryFolder;
import static org.fortiss.tooling.ext.reuse.storage.ReuseLibraryStorageManager.saveReuseLibrary;
import static org.fortiss.tooling.ext.reuse.utils.ReuseLibraryModelElementFactory.createReuseElementSpec;
import static org.fortiss.tooling.ext.reuse.utils.ReuseLibraryModelElementFactory.createReuseLibrary;
import static org.fortiss.tooling.kernel.utils.EcoreUtils.copy;
import static org.fortiss.tooling.kernel.utils.EcoreUtils.getAllReferencesSetting;
import static org.fortiss.tooling.kernel.utils.EcoreUtils.replaceEObjectReferences;
import static org.fortiss.tooling.kernel.utils.LoggingUtils.error;
import static org.fortiss.tooling.kernel.utils.LoggingUtils.showError;
import static org.fortiss.tooling.kernel.utils.UniqueIDUtils.generateMissingIDs;
import static org.fortiss.tooling.kernel.utils.UniqueIDUtils.getLargestID;
import static org.fortiss.tooling.kernel.utils.UniqueIDUtils.removeDuplicateIds;

import java.io.IOException;
import java.time.ZonedDateTime;
import java.util.ArrayList;
import java.util.Comparator;
import java.util.List;

import org.eclipse.core.runtime.CoreException;
import org.eclipse.emf.common.util.EList;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.EStructuralFeature.Setting;
import org.fortiss.tooling.base.model.element.IConnection;
import org.fortiss.tooling.base.model.element.IConnector;
import org.fortiss.tooling.base.model.element.IHierarchicElement;
import org.fortiss.tooling.base.model.element.IModelElement;
import org.fortiss.tooling.ext.reuse.model.ReuseElementSpec;
import org.fortiss.tooling.ext.reuse.model.ReuseLibrary;
import org.fortiss.tooling.ext.reuse.service.IReuseProviderService;

/**
 * Utility methods for {@link ReuseLibrary}s and related reuse elements and
 * actions where major manipulations play a role (like creations, removals,
 * updates).
 *
 * @author bergemann
 */
public class ReuseLibraryUtilsManipulation {

	/**
	 * Fixes the IDs of the model containing the given {@link EObject}, since
	 * updating reuse elements can produce duplicated IDs.
	 *
	 * @param element A reuse element of the model
	 */
	public static void fixIDs(EObject element) {
		EObject root = element;
		while (root.eContainer() != null) {
			root = root.eContainer();
		}
		removeDuplicateIds(root);
		generateMissingIDs(root, getLargestID(root));
	}

	/**
	 * Creates a hash code for the given {@link EObject} as reuse element and
	 * returns it as {@link String}. This has code format is used for detecting
	 * changes between reuse elements with the same origin. Returns null if given
	 * element is invalid (e.g. not a reuse element).
	 *
	 * @param element    The element of which a hash code should be created
	 * @param modifyDate The date of the last update/modification of the given
	 *                   element
	 * @return String of the created hash code integer (or null for an invalid
	 *         input)
	 */
	public static String createReuseHashAsString(EObject element, ZonedDateTime modifyDate) {
		if (element != null && modifyDate != null) {
			// hashCode() does not change through element differences caused by simple
			// changes like name or new ports, etc.
			// Therefore, we are currently modifying the hash with the date of the last
			// update to be sure that the hash is different after a reuse update (to see
			// e.g. if the original reuse element in the library was updated by someone else
			// in the meantime).
			int hash = element.hashCode();
			String hashString = String.valueOf(hash) + "__"
					+ ReuseLibraryUtilsBasics.getDateAsString(modifyDate, false);
			return hashString;
		}
		return null;
	}

	/**
	 * Removes all {@link IConnection}s from/to the target element except the
	 * {@link IConnection} that belongs to a {@link IConnector} of the target
	 * element that can be matched to a {@link IConnector} of the source element. In
	 * this case, the {@link IConnection} will be updated to connect now to the
	 * source element instead of the target element. This method is useful when the
	 * target element should be replaced by the source element during reuse.
	 * Regarding the matching, a {@link Comparator} must be given that realizes a
	 * compare() method that returns 0 if both given {@link IConnector}s are
	 * matching, otherwise +/-1.
	 *
	 * @param target                   The target element (e.g. the replaced
	 *                                 element)
	 * @param source                   The source element (e.g. the replacing
	 *                                 element)
	 * @param connectorMatchComparator A comparator that will return 0 with
	 *                                 compare() if both given {@link IConnector}s
	 *                                 are matching
	 */
	public static void removeNonMatchingConnections(IHierarchicElement target, IHierarchicElement source,
			Comparator<IConnector> connectorMatchComparator) {

		IConnector matchedConnector = null;
		for (IConnector targetConnector : target.getConnectors()) {
			// check if a matching connector can be found
			boolean saveConnections = false;
			for (IConnector sourceConnector : source.getConnectors()) {
				// matching is done via custom compare method (individual comparators possible)
				if (connectorMatchComparator.compare(targetConnector, sourceConnector) == 0) {
					matchedConnector = sourceConnector;
					saveConnections = true;
					replaceEObjectReferences(targetConnector, matchedConnector);
					break;
				}
			}

			// remove all connections or replace one end of them if a match was found
			List<IConnection> connectionsToBeRemoved = new ArrayList<IConnection>();
			for (IConnection incoming : targetConnector.getIncoming()) {
				if (saveConnections) {
					incoming.setTarget(matchedConnector);
				} else {
					connectionsToBeRemoved.add(incoming);
				}
			}
			for (IConnection outgoing : targetConnector.getOutgoing()) {
				if (saveConnections) {
					outgoing.setSource(matchedConnector);
				} else {
					connectionsToBeRemoved.add(outgoing);
				}
			}
			for (IConnection connection : connectionsToBeRemoved) {
				delete(connection);
			}

			// delete all external references
			if (!saveConnections) {
				for (Setting setting : getAllReferencesSetting(targetConnector)) {
					delete(setting.getEObject());
				}
			}
		}
	}

	/**
	 * Setups a given reuse element for its placement inside a {@link ReuseLibrary}.
	 *
	 * @param element        The target reuse element
	 * @param connectingSpec The reuse specification used as link/trace/reference
	 */
	public static void setupElementForLibrary(IModelElement element, ReuseElementSpec connectingSpec) {
		// currently, it is needed that only the original specification for this library
		// should exist (and also only the newest one) -> remove all others
		for (ReuseElementSpec spec : ReuseLibraryUtilsBasics.getAllReuseSpecs(element)) {
			if (!(connectingSpec.getElementUUID().equals(spec.getElementUUID())
					&& connectingSpec.getLastUpdate().equals(spec.getLastUpdate()))) {
				element.getSpecifications().remove(spec);
			}
		}
	}

	/**
	 * Creates a reuse element based on a given {@link EObject} for its placement
	 * inside a {@link ReuseLibrary}.
	 *
	 * @param originalElement The original element that should be used as base for
	 *                        the new reuse element
	 * @param library         The reuse library in which the new reuse element
	 *                        should be added (later)
	 * @param saveName        The name of the new reuse element with which it should
	 *                        be saved in the library
	 *
	 * @return The created reuse element for the library
	 */
	public static EObject createElementForLibrary(EObject originalElement, ReuseLibrary library, String saveName) {
		// add to the original element a reuse reference to the new library
		// (cast is safe: elements within the ReuseElementList are always IModelElement
		// by design)
		ReuseElementSpec spec = ReuseLibraryUtilsManipulation.addReuseSpecToElement((IModelElement) originalElement,
				library);
		spec.setElementName(saveName);
		ReuseLibraryUtilsManipulation.fixIDs(originalElement);

		// copy, because origin should still exist further (but change name of copy)
		EObject newElementForLibrary = copy(originalElement);
		ReuseLibraryUtilsBasics.setReuseElementName(newElementForLibrary, saveName);
		setupElementForLibrary((IModelElement) newElementForLibrary, spec);

		return newElementForLibrary;
	}

	/**
	 * Adds a {@link ReuseElementSpec} to the given element pointing to the given
	 * {@link ReuseLibrary} (as source).
	 *
	 * @param element The target reuse element
	 * @param library The connected reuse library
	 * @return The created reuse specification
	 */
	public static ReuseElementSpec addReuseSpecToElement(IModelElement element, ReuseLibrary library) {
		ReuseElementSpec spec = createReuseElementSpec(element, library);
		element.getSpecifications().add(spec);
		return spec;
	}

	/**
	 * Adds a given element to the given {@link ReuseLibrary}.
	 *
	 * @param element The reuse element that should be added
	 * @param library The target reuse library
	 * @throws IOException
	 * @throws CoreException
	 */
	public static void addElementToLibrary(EObject element, ReuseLibrary library) throws IOException, CoreException {
		library.getReuseElementList().add(element);
		saveReuseLibrary(library);
	}

	/**
	 * Creates a new {@link ReuseLibrary} with the given name. If no name is
	 * received, the default library name will be used.
	 *
	 * @param libraryName The name of the new reuse library
	 * @return The requested reuse library
	 */
	public static ReuseLibrary createAndAddReuseLibrary(String libraryName) {
		if (libraryName == null || libraryName.equals("")) {
			return createReuseLibrary();
		} else {
			return createReuseLibrary(libraryName);
		}
	}

	/**
	 * Replaces the given target/original reuse element in the given
	 * {@link ReuseLibrary} with the given new reuse element (which should have the
	 * given name in the end). The {@link ReuseElementSpec} references the original
	 * element (with its library).
	 *
	 * @param originalElementInLibrary The target reuse element inside a reuse
	 *                                 library that should be replace by the new
	 *                                 reuse element
	 * @param newElement               The new reuse element which is the
	 *                                 replacement of the current element in the
	 *                                 given library
	 * @param referencingSpec          The reuse specification that points to the
	 *                                 original reuse element in the library
	 * @param newName                  The new name of the element after replacement
	 */
	public static void replaceElementInLibrary(EObject originalElementInLibrary, EObject newElement,
			ReuseElementSpec referencingSpec, String newName) {
		// first, the information of the reference and the new element
		// needs to be updated and set up
		ZonedDateTime now = ZonedDateTime.now();
		referencingSpec.setLastUpdate(now);
		referencingSpec.setElementName(newName);
		referencingSpec.setElementHash(ReuseLibraryUtilsManipulation.createReuseHashAsString(newElement, now));

		EObject newLibraryElement = copy(newElement);
		ReuseLibraryUtilsBasics.setReuseElementName(newLibraryElement, newName);
		setupElementForLibrary((IModelElement) newLibraryElement, referencingSpec);

		replace(originalElementInLibrary, newLibraryElement);
	}

	/**
	 * Replaces the given (reuse) element inside its project with the given
	 * (original) reuse element from a {@link ReuseLibrary} (the element should have
	 * the given name in the end). Returns the new element in the project if the
	 * replacement/update was successful, otherwise null.
	 *
	 * @param elementInProject             The new reuse element which is the
	 *                                     replacement of the current element in the
	 *                                     given library
	 * @param originalElementInLibrary     The target reuse element inside a reuse
	 *                                     library that should be replace by the new
	 *                                     reuse element
	 * @param newName                      The new name of the element after
	 *                                     replacement
	 * @param overrideAdditionalReferences Flag whether all other possibly existing
	 *                                     references in the (old) project element
	 *                                     should be overridden/deleted (true) or if
	 *                                     they should be kept (false)
	 *
	 * @return The new element in the project if the replacement/update was
	 *         successful, otherwise null.
	 */
	public static EObject replaceElementInProject(EObject elementInProject, EObject originalElementInLibrary,
			String newName, boolean keepAdditionalReferences) {
		if (originalElementInLibrary instanceof IModelElement && elementInProject instanceof IModelElement) {
			IModelElement sourceElement = copy((IModelElement) originalElementInLibrary);
			IModelElement modelElementInProject = (IModelElement) elementInProject;
			// first, the information of the reference and the new element
			// needs to be updated and set up
			List<ReuseElementSpec> specs = ReuseLibraryUtilsBasics.getAllReuseSpecs(sourceElement);
			ZonedDateTime now = ZonedDateTime.now();
			if (specs.size() == 1) {
				// get(0) is safe due to size check
				specs.get(0).setLastUpdate(now);
			} else {
				return null;
			}
			ReuseLibraryUtilsBasics.setReuseElementName(sourceElement, newName);

			// if additional references should not be overridden but kept, they need to be
			// saved/added in the source element (otherwise they will be overridden anyway)
			if (keepAdditionalReferences) {
				List<ReuseElementSpec> oldSpecs = ReuseLibraryUtilsBasics.getAllReuseSpecs(modelElementInProject);
				// remove the specification of the current update/replacement (only additional
				// ones added)
				ReuseElementSpec mainSpec = ReuseLibraryUtilsBasics
						.getReuseSpecLinkedToThisLibraryElement(modelElementInProject, sourceElement);
				oldSpecs.remove(mainSpec);
				sourceElement.getSpecifications().addAll(oldSpecs);
			}

			boolean success = IReuseProviderService.getInstance().updateReuseElement(modelElementInProject,
					sourceElement);
			if (success) {
				return sourceElement;
			}
		}
		return null;
	}

	/**
	 * Deletes the given {@link ReuseLibrary}.
	 *
	 * @param library The target reuse library that should be deleted
	 */
	public static void deleteReuseLibrary(ReuseLibrary library) {
		// currently, everything is based only on the reuse folder/file structure, which
		// is why it is enough to just delete the local folder
		try {
			deleteReuseLibraryFolder(library);
		} catch (CoreException e) {
			error(getDefault(), "Error during deleting reuse library folder", e);
			showError("Cannot delete reuse library: " + ReuseLibraryUtilsBasics.getReuseElementName(library));
		}
	}

	/**
	 * Deletes the given reuse element by deleting its entry in the local file of
	 * the given {@link ReuseLibrary} (if it exists there).
	 *
	 * @param element       The target reuse element that should be deleted
	 * @param sourceLibrary The reuse library in which the deletion should happen
	 */
	public static void deleteReuseElementInLibraryFile(EObject element, ReuseLibrary sourceLibrary) {
		// the objects might not be the same anymore, which is why we need to identify
		// the correct object via the reuse UUID
		EObject correctLibraryElement = null;
		EList<EObject> libraryElementList = sourceLibrary.getReuseElementList();
		for (EObject libraryElement : libraryElementList) {
			if (ReuseLibraryUtilsBasics.getFirstReuseElementUUID(libraryElement)
					.equals(ReuseLibraryUtilsBasics.getReuseElementUUIDForLibrary(element, sourceLibrary))) {
				correctLibraryElement = libraryElement;
				break;
			}
		}
		if (correctLibraryElement == null) {
			return;
		}
		libraryElementList.remove(correctLibraryElement);

		try {
			saveReuseLibrary(sourceLibrary);
		} catch (IOException | CoreException e) {
			error(getDefault(), "Error during removing reuse element from reuse library", e);
			showError("Cannot remove reuse element: " + ReuseLibraryUtilsBasics.getReuseElementName(element)
					+ " from this reuse library: " + ReuseLibraryUtilsBasics.getReuseElementName(sourceLibrary));
		}
	}

	/**
	 * Deletes the given reuse element by deleting its entry in the local file of
	 * the {@link ReuseLibrary} that is mentioned as first (or only) source library
	 * for this reuse element (in its {link ReuseElementSpec}s).
	 *
	 * @param element The target reuse element that should be deleted
	 */
	public static void deleteReuseElementInLibraryFile(EObject element) {
		ReuseLibrary sourceLibrary = ReuseLibraryUtilsBasics.getFirstSourceLibraryOfElement(element);
		if (sourceLibrary != null) {
			deleteReuseElementInLibraryFile(element, sourceLibrary);
		}
	}

	/**
	 * Updates the name directly in the file of the given reuse element, which can
	 * also be a {@link ReuseLibrary}, with the given {link String}. If it is an
	 * actual reuse element (not a {@link ReuseLibrary}), the given source
	 * {@link ReuseLibrary} is where the reuse element will be updated with the new
	 * name.
	 *
	 * @param element       The target reuse element that should be updated
	 * @param newName       The new name as string
	 * @param sourceLibrary The reuse library in which the update should happen
	 */
	public static void updateReuseElementFileWithName(EObject element, String newName, ReuseLibrary sourceLibrary) {
		if (element == null || newName == null || newName.isBlank()) {
			return;
		}

		boolean modifiedLibrary = false;
		ReuseLibrary libraryForDeletion = null;
		String errorObject = "";
		if (element instanceof ReuseLibrary) {
			// since the folder and file names of a reuse library contain the library name
			// (beside its UUID), the easiest way to rename it is to save it as new library
			// (where only the name is changed, not UUID, etc.) and delete its previous
			// library folder (save and deletion are done at the end of this method)
			sourceLibrary = (ReuseLibrary) element;
			// copy, because origin should still exist further for deletion
			libraryForDeletion = copy(sourceLibrary);

			ReuseLibraryUtilsBasics.setReuseElementName(sourceLibrary, newName);
			for (EObject libraryElement : sourceLibrary.getReuseElementList()) {
				// cast is safe: elements within the ReuseElementList are always IModelElement
				// by design
				List<ReuseElementSpec> specList = ReuseLibraryUtilsBasics
						.getAllReuseSpecs((IModelElement) libraryElement);
				for (ReuseElementSpec reuseSpec : specList) {
					reuseSpec.setSourceLibName(newName);
				}
			}

			errorObject = "reuse library '" + ReuseLibraryUtilsBasics.getReuseElementName(sourceLibrary) + "'";
			modifiedLibrary = true;
		} else {
			// the objects might not be the same anymore, which is why we need to identify
			// the correct object via the reuse UUID
			for (EObject libraryElement : sourceLibrary.getReuseElementList()) {
				if (ReuseLibraryUtilsBasics.getFirstReuseElementUUID(libraryElement)
						.equals(ReuseLibraryUtilsBasics.getReuseElementUUIDForLibrary(element, sourceLibrary))) {
					ReuseLibraryUtilsBasics.setReuseElementName(libraryElement, newName);
					// cast is safe: elements within the ReuseElementList are always IModelElement
					// by design
					List<ReuseElementSpec> specList = ReuseLibraryUtilsBasics
							.getAllReuseSpecs((IModelElement) libraryElement);
					for (ReuseElementSpec reuseSpec : specList) {
						reuseSpec.setElementName(newName);
					}
					ReuseLibraryUtilsBasics.setReuseElementLastUpdateNow(libraryElement);

					errorObject = "reuse element '" + ReuseLibraryUtilsBasics.getReuseElementName(libraryElement)
							+ "' in reuse library '" + ReuseLibraryUtilsBasics.getReuseElementName(sourceLibrary) + "'";
					modifiedLibrary = true;
					break;
				}
			}
		}

		if (modifiedLibrary) {
			try {
				saveReuseLibrary(sourceLibrary);
				if (libraryForDeletion != null) {
					deleteReuseLibraryFolder(libraryForDeletion);
				}
			} catch (IOException | CoreException e) {
				error(getDefault(), "Error during renaming " + errorObject, e);
				showError("Cannot rename " + errorObject);
			}
		}
	}

	/**
	 * Updates the name directly in the file of the given reuse element, which can
	 * also be a {@link ReuseLibrary}, with the given {link String}. If it is an
	 * actual reuse element (not a {@link ReuseLibrary}), its first (or only) source
	 * {@link ReuseLibrary}, which is mentioned in its {link ReuseElementSpec}s, is
	 * where the reuse element will be updated with the new name.
	 *
	 * @param element The target reuse element that should be updated
	 * @param newName The new name as string
	 */
	public static void updateReuseElementFileWithName(EObject element, String newName) {
		if (element instanceof ReuseLibrary) {
			updateReuseElementFileWithName(element, newName, null);
		} else {
			ReuseLibrary sourceLibrary = ReuseLibraryUtilsBasics.getFirstSourceLibraryOfElement(element);
			if (sourceLibrary != null) {
				updateReuseElementFileWithName(element, newName, sourceLibrary);
			}
		}
	}

	/**
	 * Updates the comment directly in the file of the given reuse element, which
	 * can also be a {@link ReuseLibrary}, with the given {link String}. If it is an
	 * actual reuse element (not a {@link ReuseLibrary}), the given source
	 * {@link ReuseLibrary} is where the reuse element will be updated with the new
	 * comment.
	 *
	 * @param element       The target reuse element that should be updated
	 * @param newComment    The new comment as string
	 * @param sourceLibrary The reuse library in which the update should happen
	 */
	public static void updateReuseElementFileWithComment(EObject element, String newComment,
			ReuseLibrary sourceLibrary) {
		boolean modifiedLibrary = false;
		String errorObject = "";
		if (element instanceof ReuseLibrary) {
			sourceLibrary = (ReuseLibrary) element;
			ReuseLibraryUtilsBasics.setReuseElementComment(sourceLibrary, newComment);

			errorObject = "reuse library '" + ReuseLibraryUtilsBasics.getReuseElementName(sourceLibrary) + "'";
			modifiedLibrary = true;
		} else {
			// the objects might not be the same anymore, which is why we need to identify
			// the correct object via the reuse UUID
			EList<EObject> libraryElementList = sourceLibrary.getReuseElementList();
			for (EObject libraryElement : libraryElementList) {
				if (ReuseLibraryUtilsBasics.getFirstReuseElementUUID(libraryElement)
						.equals(ReuseLibraryUtilsBasics.getReuseElementUUIDForLibrary(element, sourceLibrary))) {
					ReuseLibraryUtilsBasics.setReuseElementComment(libraryElement, newComment);
					ReuseLibraryUtilsBasics.setReuseElementLastUpdateNow(libraryElement);

					errorObject = "reuse element '" + ReuseLibraryUtilsBasics.getReuseElementName(libraryElement)
							+ "' in reuse library '" + ReuseLibraryUtilsBasics.getReuseElementName(sourceLibrary) + "'";
					modifiedLibrary = true;
					break;
				}
			}
		}

		if (modifiedLibrary) {
			try {
				saveReuseLibrary(sourceLibrary);
			} catch (IOException | CoreException e) {
				error(getDefault(), "Error during modifying " + errorObject, e);
				showError("Cannot modify " + errorObject);
			}
		}
	}

	/**
	 * Updates the comment directly in the file of the given reuse element, which
	 * can also be a {@link ReuseLibrary}, with the given {link String}. If it is an
	 * actual reuse element (not a {@link ReuseLibrary}), its first (or only) source
	 * {@link ReuseLibrary}, which is mentioned in its {link ReuseElementSpec}s, is
	 * where the reuse element will be updated with the new comment.
	 *
	 * @param element    The target reuse element that should be updated
	 * @param newComment The new comment as string
	 */
	public static void updateReuseElementFileWithComment(EObject element, String newComment) {
		if (element instanceof ReuseLibrary) {
			updateReuseElementFileWithComment(element, newComment, null);
		} else {
			ReuseLibrary sourceLibrary = ReuseLibraryUtilsBasics.getFirstSourceLibraryOfElement(element);
			if (sourceLibrary != null) {
				updateReuseElementFileWithComment(element, newComment, sourceLibrary);
			}
		}
	}

}
