/*--------------------------------------------------------------------------+
|                                                                          |
| Copyright 2021 fortiss GmbH                     |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.tooling.ext.variability.ui.properties;

import static org.eclipse.swt.SWT.NONE;
import static org.fortiss.tooling.ext.variability.util.VariabilityUtils.getPresenceCondition;
import static org.fortiss.tooling.ext.variability.util.VariabilityUtils.setPresenceCondition;
import static org.fortiss.tooling.kernel.ui.util.WidgetsFactory.createTextWithUndo;

import org.eclipse.swt.events.ModifyEvent;
import org.eclipse.swt.events.ModifyListener;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Label;
import org.eclipse.swt.widgets.Text;
import org.eclipse.ui.views.properties.tabbed.TabbedPropertySheetPage;
import org.fortiss.tooling.base.model.element.IModelElement;
import org.fortiss.tooling.base.model.element.IModelElementSpecification;
import org.fortiss.tooling.base.ui.fieldassist.FieldAssist;
import org.fortiss.tooling.ext.variability.ui.util.PresenceConditionProposalProvider;
import org.fortiss.tooling.kernel.ui.extension.base.PropertySectionBase;
import org.fortiss.variability.model.presence.PresenceCondition;

/**
 * Property section for elements that may contain {@link PresenceCondition}s in
 * {@link IModelElementSpecification}s.
 * 
 * @author bayha
 */
public class HasPresenceConditionPropertySectionBase extends PropertySectionBase {

	/** Field assist for the presence condition text field. */
	protected FieldAssist pcFieldAssist;

	/** The {@link IModelElement} that this property section represents. */
	private IModelElement inputElem;

	/** The {@link Text} widget to display/edit the presence condition. */
	private Text pcText;

	/** The {@link Label} widget to label the presence condition Text. */
	private Label pcLabel;

	/** {@inheritDoc} */
	@Override
	protected void setSectionInput(Object input) {
		if(input instanceof IModelElement) {
			this.inputElem = (IModelElement)input;
			PresenceCondition pc = getPresenceCondition(inputElem);
			if(pc != null) {
				pcText.setText(pc.getStringRepresentation());
			} else {
				pcText.setText("");
			}
			createPcFieldAssist();
		} else {
			this.inputElem = null;
		}
	}

	/** Creates and sets up the field assist. */
	protected void createPcFieldAssist() {
		if(pcText != null && inputElem != null) {
			if(pcFieldAssist != null) {
				pcFieldAssist.setProposalProvider(new PresenceConditionProposalProvider(inputElem));
			} else {
				pcFieldAssist =
						new FieldAssist(pcText, new PresenceConditionProposalProvider(inputElem));
			}
		}
	}

	/** {@inheritDoc} */
	@Override
	public void createControls(Composite parent, TabbedPropertySheetPage tabbedPropertySheetPage) {
		super.createControls(parent, tabbedPropertySheetPage);

		pcText = createTextWithUndo(getWidgetFactory(), composite, "", NONE);
		pcLabel = createFormEntry(pcText, getPresenceConditionTextLabel());

		createPcFieldAssist();

		pcText.addModifyListener(new ModifyListener() {

			@Override
			public void modifyText(ModifyEvent e) {
				setPresenceCondition(inputElem, pcText.getText());
			}
		});

		pcLabel.setVisible(false);
		pcText.setVisible(false);
		pcText.setEnabled(false);
	}

	/**
	 * Retrieves the {@link String} that is displayed in fron to the presence
	 * condition text field.
	 * 
	 * @return {@link String} to be displayed in front of the presence condition
	 *         text.
	 */
	protected String getPresenceConditionTextLabel() {
		return "Presence Condition";
	}

	/** {@inheritDoc} */
	@Override
	public void refresh() {
		super.refresh();

		if(inputElem != null) {
			pcLabel.setVisible(true);
			pcText.setVisible(true);
			pcText.setEnabled(true);

			PresenceCondition pc = getPresenceCondition(inputElem);
			if(pc != null) {
				pcText.setText(pc.getStringRepresentation());
			}

		} else {
			pcText.setEnabled(false);
		}
	}
}
