/*-------------------------------------------------------------------------+
| Copyright 2022 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.tooling.ext.variability.model;

import static org.eclipse.emf.ecore.util.EcoreUtil.delete;
import static org.fortiss.tooling.ext.variability.model.VariabilityModelElementFactory.createDeactivationSpecification;

import org.eclipse.emf.common.util.BasicEList;
import org.eclipse.emf.common.util.EList;
import org.eclipse.emf.ecore.EObject;
import org.fortiss.tooling.base.model.element.IModelElement;
import org.fortiss.tooling.ext.variability.service.IDependingElementProviderService;
import org.fortiss.variability.model.IBindingStrategy;
import org.fortiss.variability.model.features.configuration.VariantConfiguration;
import org.fortiss.variability.model.presence.ICompleteEvaluationContext;
import org.fortiss.variability.model.presence.PresenceCondition;

/**
 * Static implementation for variability in AF3.
 * 
 * @author bayha
 */
public class VariabilityStaticImpl {

	/**
	 * Implements the deactivation of elements for the
	 * {@link DeactivationBindingStrategy}.
	 * 
	 * @param element
	 *            The element to be deactivated.
	 * 
	 * @return The deactivated element.
	 */
	public static <E> E deactivateElement(E element) {
		if(element instanceof IModelElement) {
			DeactivationSpecification dSpec = createDeactivationSpecification();
			((IModelElement)element).addSpecification(dSpec);
		}

		if(element instanceof EObject) {
			for(EObject e : ((EObject)element).eContents()) {
				deactivateElement(e);
			}
		}

		return element;
	}

	/**
	 * Implements the bind function for elements with a
	 * {@link OptionalVariationPointSpecification} in AF3. I.e. this is where the
	 * variability for optional elements in AF3 is bound.
	 * 
	 * @param varSpec
	 *            The {@link OptionalVariationPointSpecification} to be bound.
	 * @param context
	 *            An {@link ICompleteEvaluationContext} that enables to evaluate
	 *            the literals in varSpec. Usually this will be a
	 *            {@link VariantConfiguration}.
	 * @param strategy
	 *            The {@link IBindingStrategy} to be used.
	 * 
	 * @return The {@link IModelElement} that contained varSpec with bound
	 *         variability.
	 */
	public static IModelElement bindOptionalVariationPointSpecification(
			OptionalVariationPointSpecification varSpec, ICompleteEvaluationContext context,
			IBindingStrategy strategy) {
		PresenceCondition pc = varSpec.getPresenceCondition();
		final IModelElement modelElement = (IModelElement)varSpec.eContainer();

		if(pc == null || pc.evaluate(context)) {
			// Here, the element is supposed to be in the variant... We only need to remove
			// the presence condition, i.e. varSpec.
			delete(varSpec);
			return modelElement;
		}

		// Deactivate the depending elements first.
		varSpec.getDependingElements().stream().forEach(e -> strategy.doDeactivateElement(e));

		// Then deactivate the element itself and remove.
		return strategy.doDeactivateElement(modelElement);
	}

	/**
	 * Retrieves the {@link EObject}s which depend on the given variation point.
	 * 
	 * @param opt
	 *            The {@link OptionalVariationPointSpecification} to get the depending elements for.
	 * @return {@link EList} of depending elements.
	 */
	public static EList<EObject> getDependingElements(OptionalVariationPointSpecification opt) {
		EList<EObject> ret = new BasicEList<EObject>();
		ret.addAll(IDependingElementProviderService.getInstance()
				.getDependingElements(opt.eContainer()));
		return ret;
	}
}
