/*-------------------------------------------------------------------------+
| Copyright 2023 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.tooling.kernel.internal;

import static java.util.Collections.emptyList;
import static java.util.Collections.unmodifiableMap;
import static org.fortiss.tooling.kernel.utils.EcoreUtils.getParentsWithType;

import java.util.ArrayList;
import java.util.Collections;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;

import org.eclipse.emf.ecore.EObject;
import org.fortiss.tooling.kernel.extension.IConstraintChecker;
import org.fortiss.tooling.kernel.extension.data.IConstraintViolation;
import org.fortiss.tooling.kernel.introspection.IIntrospectionDetailsItem;
import org.fortiss.tooling.kernel.introspection.IIntrospectionItem;
import org.fortiss.tooling.kernel.introspection.IIntrospectiveKernelService;
import org.fortiss.tooling.kernel.introspection.items.ConstraintCheckerKISSDetailsItem;
import org.fortiss.tooling.kernel.service.IConstraintCheckerService;
import org.fortiss.tooling.kernel.service.IKernelIntrospectionSystemService;
import org.fortiss.tooling.kernel.service.base.EObjectAwareServiceBase;

/**
 * This class implements the {@link IConstraintCheckerService} interface.
 * 
 * @author hoelzl
 */
public class ConstraintCheckerService extends EObjectAwareServiceBase<IConstraintChecker<EObject>>
		implements IConstraintCheckerService, IIntrospectiveKernelService {
	/** The singleton instance. */
	private static final ConstraintCheckerService INSTANCE = new ConstraintCheckerService();

	/** Returns singleton instance of the service. */
	public static ConstraintCheckerService getInstance() {
		return INSTANCE;
	}

	/** The constraint checker extension point ID. */
	private static final String EXTENSION_POINT_NAME =
			"org.fortiss.tooling.kernel.modelElementConstraintChecker";

	/** The constraint checker configuration element name. */
	private static final String CONFIGURATION_ELEMENT_NAME = "modelElementConstraintChecker";

	/** The constraint checker handler attribute name. */
	private static final String HANDLER_CLASS_ATTRIBUTE_NAME = "checker";

	/**
	 * A list of all the classes whose elements are excluded from constraint checks. This exclusion
	 * also applies to all of their children, i.e., if an element has a parent whose class is stored
	 * in this list, it will also be excluded from all constraint checks.
	 */
	private static List<Class<? extends EObject>> constraintCheckExclusionTypes = new ArrayList<>();

	/** {@inheritDoc} */
	@Override
	public void startService() {
		IKernelIntrospectionSystemService.getInstance().registerService(this);
	}

	/** {@inheritDoc} */
	@Override
	public void registerConstraintChecker(IConstraintChecker<EObject> checker,
			Class<?> modelElementClass) {
		addHandler(modelElementClass, checker);
	}

	/** {@inheritDoc} */
	@Override
	public void registerTypeAsExcludedParentForConstraintChecks(Class<? extends EObject> clazz) {
		constraintCheckExclusionTypes.add(clazz);
	}

	/** {@inheritDoc} */
	@Override
	public String getIntrospectionDescription() {
		return getIntrospectionLabel() +
				"\n\nThis service allows the registration of constraint checkers for models and/or model elements." +
				"\nConstraint checks are performed on model elements and may create constraint violations of different severities." +
				"\nViolations are usually displayed to the tool user and require some action to clear it." +
				"\n\nThe service extension point is '" + EXTENSION_POINT_NAME + "'.";
	}

	/** {@inheritDoc} */
	@Override
	public List<IConstraintViolation<? extends EObject>>
			performAllConstraintChecksRecursively(EObject modelElement) {
		List<IConstraintViolation<? extends EObject>> result =
				new LinkedList<IConstraintViolation<? extends EObject>>();
		performConstraintCheck(modelElement, result);
		for(Iterator<EObject> iter = modelElement.eAllContents(); iter.hasNext();) {
			performConstraintCheck(iter.next(), result);
		}
		Collections.sort(result, IConstraintViolation.SEVERITY_COMPARATOR);
		return result;
	}

	/** {@inheritDoc} */
	@Override
	public List<IConstraintViolation<? extends EObject>>
			performAllConstraintChecks(EObject modelElement) {
		List<IConstraintViolation<? extends EObject>> result =
				new LinkedList<IConstraintViolation<? extends EObject>>();
		performConstraintCheck(modelElement, result);
		Collections.sort(result, IConstraintViolation.SEVERITY_COMPARATOR);
		return result;
	}

	/**
	 * Performs all constraint checks on the given model element if itself or its parents are not
	 * excluded from such checks. Violations are added to the given list. Constraint checks on
	 * content elements are not considered.
	 */
	private void performConstraintCheck(EObject modelElement,
			List<IConstraintViolation<? extends EObject>> violationList) {

		// Skip constraint checks if the given element itself is excluded from checks or if it
		// exists within an excluded element.
		// TODO (TM): Looking through all parents for each excluded type introduces quite an
		// overhead. As the list of parents is never used, memory and runtime overhead can be
		// reduced by using `getFirstParentWithType` or, even better, `isAncestor` instead.
		for(Class<? extends EObject> excludedClass : constraintCheckExclusionTypes) {
			List<? extends EObject> foundExcludedParents =
					getParentsWithType(modelElement, excludedClass);
			Class<?> givenClass = modelElement.getClass();
			if(givenClass.isAssignableFrom(excludedClass) || !foundExcludedParents.isEmpty()) {
				return;
			}
		}

		List<IConstraintChecker<EObject>> handlers = getRegisteredHandlers(modelElement.getClass());
		if(handlers == null) {
			return;
		}
		for(IConstraintChecker<EObject> checker : handlers) {
			if(checker.isApplicable(modelElement)) {
				List<? extends IConstraintViolation<? extends EObject>> violation =
						checker.apply(modelElement);
				if(violation != null) {
					violationList.addAll(violation);
				}
			}
		}
	}

	/** {@inheritDoc} */
	@Override
	public List<IConstraintChecker<? extends EObject>>
			getAllConstraintCheckers(EObject modelElement) {
		return new ArrayList<IConstraintChecker<? extends EObject>>(
				getRegisteredHandlers(modelElement.getClass()));
	}

	/** {@inheritDoc} */
	@Override
	protected String getExtensionPointName() {
		return EXTENSION_POINT_NAME;
	}

	/** {@inheritDoc} */
	@Override
	protected String getConfigurationElementName() {
		return CONFIGURATION_ELEMENT_NAME;
	}

	/** {@inheritDoc} */
	@Override
	protected String getHandlerClassAttribute() {
		return HANDLER_CLASS_ATTRIBUTE_NAME;
	}

	/** {@inheritDoc} */
	@Override
	public String getIntrospectionLabel() {
		return "Constraint Checker Service";
	}

	/** {@inheritDoc} */
	@Override
	public List<IIntrospectionItem> getIntrospectionItems() {
		return emptyList();
	}

	/** {@inheritDoc} */
	@Override
	public boolean showInIntrospectionNavigation() {
		return true;
	}

	/** {@inheritDoc} */
	@Override
	public IIntrospectionDetailsItem getDetailsItem() {
		return new ConstraintCheckerKISSDetailsItem(unmodifiableMap(handlerMap));
	}
}
