/*--------------------------------------------------------------------------+
| Copyright 2021 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.variability.model;

import static java.util.stream.Collectors.toList;
import static org.eclipse.emf.ecore.util.EcoreUtil.delete;
import static org.fortiss.tooling.kernel.utils.EcoreUtils.getChildrenWithType;
import static org.fortiss.variability.model.VariabilityModelElementFactory.createAndPC;
import static org.fortiss.variability.model.VariabilityModelElementFactory.createLiteralPC;
import static org.fortiss.variability.model.VariabilityModelElementFactory.createNotPC;

import java.util.List;
import java.util.stream.Stream;

import org.eclipse.emf.common.util.EList;
import org.eclipse.emf.ecore.EObject;
import org.fortiss.variability.model.base.RemoveBindingStrategy;
import org.fortiss.variability.model.features.AbstractAlternativeFeature;
import org.fortiss.variability.model.features.AbstractFeature;
import org.fortiss.variability.model.features.configuration.AlternativeFeatureConfiguration;
import org.fortiss.variability.model.features.configuration.CompositionalFeatureConfiguration;
import org.fortiss.variability.model.features.configuration.IFeatureConfiguration;
import org.fortiss.variability.model.presence.AndPC;
import org.fortiss.variability.model.presence.ICompleteEvaluationContext;
import org.fortiss.variability.model.presence.ILiteralReferencable;
import org.fortiss.variability.model.presence.LiteralPC;
import org.fortiss.variability.model.presence.NotPC;
import org.fortiss.variability.model.presence.OrPC;
import org.fortiss.variability.model.presence.PresenceCondition;
import org.fortiss.variability.model.presence.PresenceConditionTerm;

/**
 * Static implementations for variability model.
 * 
 * @author bayha
 */
public class VariabilityStaticImpl {

	/** Hint for selecting alternatives in the UI. */
	public static final String ALTERNATIVE_NOTHING_SELECTED_STRING = "<Select Alternative>";

	/** String representation of an empty alternative. */
	public static final String OPTIONAL_ALTERNATIVE_NOT_SELECTED_STRING = "None";

	/**
	 * Binds variability in optional variation points.
	 * 
	 * @param variationPoint
	 *            The {@link IOptionalVariationPoint} to bind
	 *            variability for.
	 * @param context
	 *            The ICompleteEvaluationContext to use for evaluating
	 *            presence conditions.
	 * @param strategy
	 *            The {@link IBindingStrategy} to use for binding
	 *            variability.
	 * @return The result of binding variability in variationPoint.
	 */
	@SuppressWarnings("unchecked")
	public static <T> T bindOptionalVariationPoint(IOptionalVariationPoint<T> variationPoint,
			ICompleteEvaluationContext context, IBindingStrategy strategy) {
		PresenceCondition pc = variationPoint.getPresenceCondition();
		if(pc == null || pc.evaluate(context)) {
			return((T)variationPoint);
		}

		return strategy.deactivateOptionalVariationPoint(variationPoint);
	}

	/**
	 * Translates the given {@link IFeatureConfiguration} and all sub configurations
	 * to a {@link PresenceConditionTerm}.
	 * 
	 * @param featureConfig
	 *            The {@link IFeatureConfiguration} to translate.
	 * @return A {@link PresenceConditionTerm}, representing {@code featureConfig}
	 *         and all sub-configurations.
	 */
	public static PresenceConditionTerm translateFeatureConfigurationToPresenceCondition(
			IFeatureConfiguration<?> featureConfig) {
		final AbstractFeature feature = featureConfig.getFeatureReference();
		final LiteralPC featurePCTerm = createLiteralPC(feature);

		if(featureConfig.isSelected()) {
			if(featureConfig instanceof CompositionalFeatureConfiguration) {
				// Conjugate and return all subFeaturesConfigurations
				final EList<IFeatureConfiguration<?>> allSubConfig =
						((CompositionalFeatureConfiguration)featureConfig).getSubConfigurations();
				final Stream<PresenceConditionTerm> allSubConfigTerms =
						allSubConfig.stream().map(sc -> sc.translateToPresenceCondition());

				return allSubConfigTerms.reduce(featurePCTerm, (a, b) -> createAndPC(a, b, ""));

			} else if(featureConfig instanceof AlternativeFeatureConfiguration) {
				// Conjugate the selected alternative's term with unselected alternative's terms
				// and the given feature configurations term
				final IFeatureConfiguration<?> selectedSubConfig =
						((AlternativeFeatureConfiguration)featureConfig).getSelectedConfiguration();
				final PresenceConditionTerm selectedSubConfigTerm =
						translateFeatureConfigurationToPresenceCondition(selectedSubConfig);

				if(feature instanceof AbstractAlternativeFeature) {
					// Collect all alternatives which were not selected
					List<AbstractFeature> alternatives =
							((AbstractAlternativeFeature)feature).getAlternatives();
					alternatives = alternatives.stream()
							.filter(f -> !f.equals(selectedSubConfig.getFeatureReference()))
							.collect(toList());

					final PresenceConditionTerm subTerms = alternatives.stream()
							.map(a -> translateFeatureToDeselectedPresenceCondition(a))
							.reduce(selectedSubConfigTerm, (a, b) -> createAndPC(a, b, ""));
					return createAndPC(featurePCTerm, subTerms, "");
				}
			}

			throw new RuntimeException("Invalid FeatureConfiguration found for feature \"" +
					feature.getName() + "\":\n" + featureConfig);

		}
		// Negate the current feature and all sub-features
		return translateFeatureToDeselectedPresenceCondition(feature);
	}

	/**
	 * Creates a {@link PresenceConditionTerm} which negates the given feature and
	 * all sub-features.
	 */
	private static PresenceConditionTerm
			translateFeatureToDeselectedPresenceCondition(final AbstractFeature feature) {
		// Create and return a recursive term "NOT feature AND NOT feature.SubFeature1
		// AND NOT feature.SubFeature2 AND ..."
		PresenceConditionTerm notCurrent = createNotPC(createLiteralPC(feature), "");
		final EList<AbstractFeature> allSubFeatures =
				getChildrenWithType(feature, AbstractFeature.class);

		final Stream<PresenceConditionTerm> allSubFeaturesNegatedTerms =
				allSubFeatures.stream().map(sf -> createNotPC(createLiteralPC(sf), ""));
		// Conjugate all negations for sub-features with current feature.
		final PresenceConditionTerm result =
				allSubFeaturesNegatedTerms.reduce(notCurrent, (a, f) -> createAndPC(a, f, ""));
		return result;
	}

	/**
	 * Translates the given {@link IFeatureConfiguration} to a String
	 * representation.
	 * 
	 * @param featureConf
	 *            The {@link IFeatureConfiguration} to be translated.
	 * @return A {@link String} representing the given {@code featureConf}
	 */
	public static String featureConfigurationValueToString(IFeatureConfiguration<?> featureConf) {
		if(featureConf instanceof AlternativeFeatureConfiguration) {
			AlternativeFeatureConfiguration afc = (AlternativeFeatureConfiguration)featureConf;
			IFeatureConfiguration<?> selectedConfiguration = afc.getSelectedConfiguration();
			if(selectedConfiguration != null) {
				return selectedConfiguration.getFeatureReference().getName();
			}
			if(afc.getFeatureReference().isOptional()) {
				return OPTIONAL_ALTERNATIVE_NOT_SELECTED_STRING;
			}
			return ALTERNATIVE_NOTHING_SELECTED_STRING;
		}

		if(featureConf.isSelected()) {
			return "Selected";
		}
		return "Deselected";
	}

	/**
	 * Evaluates the given {@code literal} for the given
	 * {@link IFeatureConfiguration}.
	 * 
	 * @param varConf
	 *            The {@link IFeatureConfiguration} to evaluate the
	 *            {@code literal for}.
	 * @param literal
	 *            The {@link LiteralPC} to be evaluated.
	 * @return 'true' if the given {@code literal} is true for the given
	 *         {@code varConf}. 'false' otherwise.
	 */
	public static boolean evaluateLiteral(IFeatureConfiguration<?> varConf, LiteralPC literal) {
		ILiteralReferencable litRef = literal.getLiteralReference();
		AbstractFeature confFeat = varConf.getFeatureReference();

		// If the literal reference is null, the feature could not be found,
		// when parsing the presence condition. It cannot be evaluated and the
		// element should not be removed. Hence, "true" is returned here.
		if(litRef == null) {
			System.err.println(
					"Empy Literal found in Presence Condition. String representation is: \"" +
							literal.getStringRepresentation() + "\". The literal will be ignored!");
			return true;
		}

		// Recursively evaluate literal for feature configuration
		if(litRef.equals(confFeat) && varConf.isSelected()) {
			return true;
		}

		if(varConf instanceof AlternativeFeatureConfiguration) {
			IFeatureConfiguration<?> selectedAltConf =
					((AlternativeFeatureConfiguration)varConf).getSelectedConfiguration();
			return selectedAltConf != null && evaluateLiteral(selectedAltConf, literal);
		}
		if(varConf instanceof CompositionalFeatureConfiguration) {
			return ((CompositionalFeatureConfiguration)varConf).getSubConfigurations().stream()
					.anyMatch(c -> evaluateLiteral(c, literal));
		}

		throw new RuntimeException("Unknown FeatureCofiguration of type " +
				varConf.eClass().getName() + " found when evaluating the literal " +
				literal.getStringRepresentation() + "!");
	}

	/**
	 * Evaluate the given {@link PresenceCondition} {@code pc} in the given
	 * evaluation {@code context}.
	 * 
	 * @param pc
	 *            The {@link PresenceCondition} to be evaluated.
	 * @param context
	 *            The {@link ICompleteEvaluationContext} to be used for
	 *            evaluating {@code pc}.
	 * @return 'true' if {@code pc} is 'true' in the given {@code context}.
	 *         'false' otherwise.
	 */
	public static boolean evaluatePresenceCondition(PresenceCondition pc,
			ICompleteEvaluationContext context) {
		if(pc instanceof LiteralPC) {
			LiteralPC lit = (LiteralPC)pc;
			return context.evaluateLiteral(lit);
		}
		if(pc instanceof NotPC) {
			NotPC not = (NotPC)pc;
			return !not.getOperand().evaluate(context);
		}
		if(pc instanceof OrPC) {
			OrPC or = (OrPC)pc;
			return or.getOperand1().evaluate(context) || or.getOperand2().evaluate(context);
		}
		if(pc instanceof AndPC) {
			AndPC and = (AndPC)pc;
			return and.getOperand1().evaluate(context) && and.getOperand2().evaluate(context);
		}

		throw new IllegalArgumentException(
				"The passed PresenceCondition is null or cannot be evaluated!");
	}

	/**
	 * Deactivates the given {@link IOptionalVariationPoint} {@code element} using
	 * the given {@code strategy}.
	 * 
	 * <D> it the type parameter of the variation type. I.e. the type it will
	 * evaluate to when binding variability.
	 * 
	 * @param element
	 *            The {@link IOptionalVariationPoint} to be deactivated.
	 * @param strategy
	 *            The {@link IBindingStrategy} to be used for deactivation.
	 * @return The result of deactivating {@code element} of type {@code D}.
	 */
	@SuppressWarnings("unchecked")
	public static <D> D deactivateOptionalVariationPoint(IOptionalVariationPoint<D> element,
			IBindingStrategy strategy) {

		EList<EObject> dependingElements = element.getDependingElements();
		dependingElements.stream().forEach(e -> strategy.doDeactivateElement(e));

		return (D)strategy.doDeactivateElement(element);
	}

	/**
	 * Deactivates the given {@code element} from the model by deleting it.
	 * 
	 * @param element
	 *            The object of type {@code E} to be removed.
	 * @return The representation of the object after deactivating it. For deletion
	 *         here, this will always be {@code null}.
	 */
	public static <E> E doRemoveElement(E element) {
		if(element instanceof EObject) {
			delete((EObject)element, true);
		}

		return null;
	}

	/**
	 * Retrieves the {@link RemoveBindingStrategy} singleton.
	 * 
	 * @return The singleton {@link RemoveBindingStrategy} object.
	 */
	public static RemoveBindingStrategy getRemoveBindingStrategyInstance() {
		return getRemoveBindingStrategyInstance();
	}

	/**
	 * Implements equality (equals) for features.
	 * 
	 * @param subject
	 *            The {@link AbstractFeature} to be checked against
	 *            {@code target}.
	 * @param target
	 *            The {@link Object} to be checked against {@code subject}.
	 * @return 'true' if {@code subject} and {@code target} represent the same
	 *         feature. 'false' otherwise.
	 */
	public static boolean areFeaturesEqual(AbstractFeature subject, Object target) {
		if(target == null) {
			return false;
		}

		return target.getClass() == subject.getClass() &&
				((AbstractFeature)target).getName().equals(subject.getName());
	}
}
